@@ -25,6 +25,7 @@ def bench(session: nox.Session) -> None:
 @nox.session(python=["3.8", "3.9", "3.10", "3.11", "pypy3.8", "pypy3.9"])
 def tests(session: nox.Session) -> None:
     args = ("--no-deps",) if os.getenv("SKIP_DEPS_INSTALL") == "true" else ()
+    session.install("--upgrade", "pip")
     session.install(".[tests]", *args)
     session.run(
         "pytest",
@@ -38,7 +39,7 @@ def tests(session: nox.Session) -> None:
 @nox.session
 def lint(session: nox.Session) -> None:
     session.install("pre-commit")
-    session.install("-e", ".[dev]")
+    session.install("-e", ".[dev,vector]")
 
     args = *(session.posargs or ("--show-diff-on-failure",)), "--all-files"
     session.run("pre-commit", "run", *args)
