@@ -5,16 +5,16 @@ import sys
 import traceback
 from argparse import Action, ArgumentParser, ArgumentTypeError, Namespace
 from itertools import chain
+from multiprocessing import freeze_support
 from typing import Iterable, Iterator, List, Mapping, Optional, Tuple, Union
 
 import shtab
 
 from dql import __version__, utils
-from dql.catalog import Catalog
+from dql.catalog import Catalog, get_catalog, indexer_formats
 from dql.cli_utils import BooleanOptionalAction
 from dql.client import Client
-from dql.data_storage import SQLiteDataStorage
-from dql.node import long_line_str
+from dql.node import DirType, long_line_str
 from dql.remote.studio import StudioClient
 from dql.utils import DQLDir, get_remote_config, read_config
 
@@ -97,24 +97,14 @@ def get_parser() -> ArgumentParser:
     parent_parser.add_argument(
         "-q", "--quiet", action="count", default=0, help="Be quiet"
     )
+    parent_parser.add_argument(
+        "--debug-sql",
+        action="store_true",
+        default=False,
+        help="Show All SQL Queries (very verbose output, for debugging only)",
+    )
 
     subp = parser.add_subparsers(help="Sub-command help", dest="command")
-    parse_get = subp.add_parser("get", parents=[parent_parser], help="Fetch a dataset")
-    parse_get.add_argument(
-        "source", type=str, help="Data source - a path to a cloud storage dir"
-    )
-    parse_get.add_argument(  # type: ignore[attr-defined]
-        "-o", "--output", type=str, help="Output"
-    ).complete = shtab.DIR
-    parse_get.add_argument(
-        "-f",
-        "--force",
-        action="count",
-        default=0,
-        help="Force creating outputs",
-    )
-    parse_get.add_argument("-d", "--descr", type=str, help="Description")
-
     parse_cp = subp.add_parser(
         "cp", parents=[parent_parser], help="Copy data files from the cloud"
     )
@@ -162,16 +152,6 @@ def get_parser() -> ArgumentParser:
         action="store_true",
         help="Copy directories recursively",
     )
-    parse_clone.add_argument(
-        "--dql-file",
-        help="Use a different filename for the resulting DQL file",
-    )
-    parse_clone.add_argument(
-        "--dql-only",
-        default=False,
-        action="store_true",
-        help="Only create the resulting DQL file, do not download or copy files",
-    )
     parse_clone.add_argument(
         "--no-glob",
         default=False,
@@ -179,10 +159,20 @@ def get_parser() -> ArgumentParser:
         help="Do not expand globs (such as * or ?)",
     )
     parse_clone.add_argument(
-        "--no-dql-file",
+        "--no-cp",
         default=False,
         action="store_true",
-        help="Do not write a .dql file",
+        help="Do not copy files, just create a shadow dataset",
+    )
+    parse_clone.add_argument(
+        "--edql",
+        default=False,
+        action="store_true",
+        help="Create a .edql file",
+    )
+    parse_clone.add_argument(
+        "--edql-file",
+        help="Use a different filename for the resulting .edql file",
     )
 
     parse_register = subp.add_parser(
@@ -278,6 +268,31 @@ def get_parser() -> ArgumentParser:
         help="Force delete registered dataset with all of it's versions",
     )
 
+    dataset_stats_parser = subp.add_parser(
+        "dataset-stats", parents=[parent_parser], help="Shows basic dataset stats"
+    )
+    dataset_stats_parser.add_argument("name", type=str, help="Dataset name")
+    dataset_stats_parser.add_argument(
+        "--version",
+        action="store",
+        default=None,
+        type=int,
+        help="Dataset version",
+    )
+    dataset_stats_parser.add_argument(
+        "-b",
+        "--bytes",
+        default=False,
+        action="store_true",
+        help="Display size in bytes instead of human-readable size",
+    )
+    dataset_stats_parser.add_argument(
+        "--si",
+        default=False,
+        action="store_true",
+        help="Display size using powers of 1000 not 1024",
+    )
+
     parse_merge_datasets = subp.add_parser(
         "merge-datasets", parents=[parent_parser], help="Merges datasets"
     )
@@ -425,6 +440,13 @@ def get_parser() -> ArgumentParser:
     parse_index = subp.add_parser(
         "index", parents=[parent_parser], help="Index storage location"
     )
+    parse_index.add_argument(
+        "--format",
+        type=str,
+        default=None,
+        choices=indexer_formats.keys(),
+        help="Postprocessing step to apply, after indexing storage.",
+    )
     add_sources_arg(parse_index)
     add_storage_parser = subp.add_parser(
         "add-storage", parents=[parent_parser], help="Register local path as storage"
@@ -442,6 +464,12 @@ def get_parser() -> ArgumentParser:
         parents=[parent_parser],
         help="Finds and marks stale storages",
     )
+    apply_udf_parser = subp.add_parser(
+        "apply-udf", parents=[parent_parser], help="Apply UDF"
+    )
+    apply_udf_parser.add_argument("udf", type=str, help="UDF location")
+    apply_udf_parser.add_argument("source", type=str, help="Source storage or dataset")
+    apply_udf_parser.add_argument("target", type=str, help="Target dataset name")
 
     add_completion_parser(subp, [parent_parser])
     return parser
@@ -470,37 +498,6 @@ def get_logging_level(args: Namespace) -> int:
     return logging.INFO
 
 
-def get(source, output, **kwargs):
-    data_storage = SQLiteDataStorage()
-    catalog = Catalog(data_storage)
-    catalog.get(source, output, **kwargs)
-
-
-def cp(sources, output, **kwargs):
-    data_storage = SQLiteDataStorage()
-    catalog = Catalog(data_storage)
-    catalog.cp(sources, output, **kwargs)
-
-
-def register(name, **kwargs):
-    data_storage = SQLiteDataStorage()
-    catalog = Catalog(data_storage)
-    catalog.register_shadow_dataset(name, **kwargs)
-    print(f"Registered dataset {name}")
-
-
-def edit_dataset(name, **kwargs):
-    data_storage = SQLiteDataStorage()
-    catalog = Catalog(data_storage)
-    catalog.edit_dataset(name, **kwargs)
-
-
-def merge_datasets(src, dst, **kwargs):
-    data_storage = SQLiteDataStorage()
-    catalog = Catalog(data_storage)
-    catalog.merge_datasets(src, dst, **kwargs)
-
-
 def ls_urls(
     sources,
     long: bool = False,
@@ -512,7 +509,7 @@ def ls_urls(
     if client_config is None:
         client_config = {}
     if catalog is None:
-        catalog = Catalog(SQLiteDataStorage())
+        catalog = get_catalog()
 
     curr_dir = None
     value_iterables = []
@@ -529,6 +526,19 @@ def ls_urls(
         yield curr_dir, chain(*value_iterables)
 
 
+def _node_data_to_ls_values(row, long_format=False):
+    name = row[0]
+    is_dir = row[1] == DirType.DIR
+    ending = "/" if is_dir else ""
+    value = name + ending
+    if long_format:
+        last_modified = row[2]
+        owner_name = row[3]
+        timestamp = last_modified if not is_dir else None
+        return long_line_str(value, timestamp, owner_name)
+    return value
+
+
 def _ls_urls_flat(
     sources,
     long: bool,
@@ -547,15 +557,15 @@ def _ls_urls_flat(
             yield source, values
         else:
             found = False
-            for data_source, nodes in catalog.ls(
-                [source], client_config=client_config, **kwargs
+            fields = ["name", "dir_type"]
+            if long:
+                fields.extend(["last_modified", "owner_name"])
+            for data_source, results in catalog.ls(
+                [source], fields=fields, client_config=client_config, **kwargs
             ):
-                if long:
-                    values = (n.long_line_str() for n in nodes)
-                else:
-                    values = (n.name_with_dir_ending for n in nodes)
+                values = (_node_data_to_ls_values(r, long) for r in results)
                 found = True
-                yield data_source.node.dirname, values
+                yield data_source.dirname(), values
             if not found:
                 raise FileNotFoundError(f"No such file or directory: {source}")
 
@@ -564,20 +574,19 @@ def ls_indexed_storages(
     long: bool = False, catalog: Optional[Catalog] = None
 ) -> Iterator[str]:
     if catalog is None:
-        catalog = Catalog(SQLiteDataStorage())
-    storages = catalog.ls_storages()
+        catalog = get_catalog()
+    storage_uris = catalog.ls_storage_uris()
     if long:
-        for s in storages:
+        for uri in storage_uris:
             # TODO: add Storage.created so it can be used here
-            yield long_line_str(s.uri, None, "")
+            yield long_line_str(uri, None, "")
     else:
-        for s in storages:
-            yield s.uri
+        yield from storage_uris
 
 
 def ls_local(sources, long: bool = False, catalog: Optional[Catalog] = None, **kwargs):
     if catalog is None:
-        catalog = Catalog(SQLiteDataStorage())
+        catalog = get_catalog()
     if sources:
         if len(sources) == 1:
             for _, entries in ls_urls(sources, long=long, catalog=catalog, **kwargs):
@@ -664,28 +673,42 @@ def ls(
 
 
 def ls_datasets(shadow_only=None):
-    for d in Catalog(SQLiteDataStorage()).ls_datasets(shadow_only=shadow_only):
+    for d in get_catalog().ls_datasets(shadow_only=shadow_only):
         if d.shadow:
             print(f"(tmp) {d.name}")
         else:
             for v in d.versions or []:
-                print(f"{d.name} (v{v})")
+                print(f"{d.name} (v{v.version})")
 
 
 def ls_dataset_rows(name: str, version: Optional[int] = None):
-    for row in Catalog(SQLiteDataStorage()).ls_dataset_rows(name, version=version):
+    for row in get_catalog().ls_dataset_rows(name, version=version):
         entry = row.name + ("/" if row.dir_type else "")  # type: ignore[attr-defined]
         print(format_ls_entry(entry))
 
 
 def rm_dataset(name: str, version: Optional[int] = None, force: Optional[bool] = False):
-    Catalog(SQLiteDataStorage()).remove_dataset(name, version=version, force=force)
+    get_catalog().remove_dataset(name, version=version, force=force)
+
+
+def dataset_stats(
+    name: str,
+    version: Optional[int] = None,
+    show_bytes=False,
+    si=False,
+):
+    stats = get_catalog().dataset_stats(name, version=version)
+
+    if stats:
+        print(f"Number of objects: {stats.num_objects}")
+        if show_bytes:
+            print(f"Total objects size: {stats.size}")
+        else:
+            print(f"Total objects size: {utils.sizeof_fmt(stats.size, si=si): >7}")
 
 
 def du(sources, show_bytes=False, si=False, **kwargs):
-    data_storage = SQLiteDataStorage()
-    catalog = Catalog(data_storage)
-    for path, size in catalog.du(sources, **kwargs):
+    for path, size in get_catalog().du(sources, **kwargs):
         if show_bytes:
             print(f"{size} {path}")
         else:
@@ -693,23 +716,33 @@ def du(sources, show_bytes=False, si=False, **kwargs):
 
 
 def find(sources, **kwargs):
-    data_storage = SQLiteDataStorage()
-    catalog = Catalog(data_storage)
-    yield from catalog.find(sources, **kwargs)
+    yield from get_catalog().find(sources, **kwargs)
 
 
-def index(sources, catalog: Optional[Catalog] = None, **kwargs):
+def index(
+    sources,
+    catalog: Optional[Catalog] = None,
+    indexer_format: Optional[str] = None,
+    **kwargs,
+):
     if catalog is None:
-        catalog = Catalog(SQLiteDataStorage())
-    catalog.index(sources, **kwargs)
+        catalog = get_catalog()
+    fmt = None
+    if indexer_format:
+        fmt = indexer_formats.get(indexer_format)
+    catalog.index(sources, index_processors=fmt, **kwargs)
 
 
 def add_storage(uri: str, symlinks: bool = False) -> None:
-    Catalog(SQLiteDataStorage()).add_storage(uri, symlinks=symlinks)
+    get_catalog().add_storage(uri, symlinks=symlinks)
 
 
 def find_stale_storages():
-    Catalog(SQLiteDataStorage()).find_stale_storages()
+    get_catalog().find_stale_storages()
+
+
+def apply_udf(udf: str, source: str, target: str):
+    get_catalog().apply_udf(udf, source, target)
 
 
 def completion(shell: str) -> str:
@@ -720,6 +753,9 @@ def completion(shell: str) -> str:
 
 
 def main(argv: Optional[List[str]] = None) -> int:  # noqa: C901
+    # Required for Windows multiprocessing support
+    freeze_support()
+
     parser = get_parser()
     args = parser.parse_args(argv)
 
@@ -736,63 +772,57 @@ def main(argv: Optional[List[str]] = None) -> int:  # noqa: C901
         "aws_anon": args.aws_anon,
     }
 
+    if args.debug_sql:
+        # This also sets this environment variable for any subprocesses
+        os.environ["DEBUG_SHOW_SQL_QUERIES"] = "True"
+
     try:
-        if args.command == "get":
-            # TODO: descr = args.descr
-            get(
-                args.source,
-                args.output,
-                force=bool(args.force),
-                update=bool(args.update),
-                ttl=args.ttl,
-                client_config=client_config,
-            )
-        elif args.command == "cp":
-            cp(
+        if args.command == "cp":
+            get_catalog().cp(
                 args.sources,
                 args.output,
                 force=bool(args.force),
                 update=bool(args.update),
                 recursive=bool(args.recursive),
-                dql_file=None,
-                dql_only=False,
+                edql_file=None,
+                edql_only=False,
+                no_edql_file=True,
                 no_glob=args.no_glob,
-                no_dql_file=True,
                 ttl=args.ttl,
                 client_config=client_config,
             )
         elif args.command == "clone":
-            cp(
+            get_catalog().clone(
                 args.sources,
                 args.output,
                 force=bool(args.force),
                 update=bool(args.update),
                 recursive=bool(args.recursive),
-                dql_file=args.dql_file,
-                dql_only=args.dql_only,
                 no_glob=args.no_glob,
-                no_dql_file=args.no_dql_file,
                 ttl=args.ttl,
+                no_cp=args.no_cp,
+                edql=args.edql,
+                edql_file=args.edql_file,
                 client_config=client_config,
-                create_dataset=True,
             )
         elif args.command == "register":
-            register(
+            get_catalog().register_shadow_dataset(
                 args.name,
                 description=args.description,
                 labels=args.labels,
                 registered_name=args.registered_name,
                 version=args.version,
             )
+            print(f"Registered dataset {args.name}")
         elif args.command == "edit-dataset":
-            edit_dataset(
+            get_catalog().edit_dataset(
                 args.name,
                 description=args.description,
                 new_name=args.new_name,
                 labels=args.labels,
             )
         elif args.command == "merge-datasets":
-            merge_datasets(
+            get_catalog().merge_datasets(
                 args.src,
                 args.dst,
                 src_version=args.src_version,
@@ -813,6 +843,13 @@ def main(argv: Optional[List[str]] = None) -> int:  # noqa: C901
             ls_dataset_rows(args.name, args.version)
         elif args.command == "rm-dataset":
             rm_dataset(args.name, version=args.version, force=args.force)
+        elif args.command == "dataset-stats":
+            dataset_stats(
+                args.name,
+                version=args.version,
+                show_bytes=args.bytes,
+                si=args.si,
+            )
         elif args.command == "du":
             du(
                 args.sources,
@@ -847,6 +884,7 @@ def main(argv: Optional[List[str]] = None) -> int:  # noqa: C901
             index(
                 args.sources,
                 ttl=args.ttl,
+                indexer_format=args.format,
                 update=bool(args.update),
                 client_config=client_config,
             )
@@ -856,6 +894,8 @@ def main(argv: Optional[List[str]] = None) -> int:  # noqa: C901
             print(completion(args.shell))
         elif args.command == "find-stale-storages":
             find_stale_storages()
+        elif args.command == "apply-udf":
+            apply_udf(args.udf, args.source, args.target)
         else:
             print(f"invalid command: {args.command}", file=sys.stderr)
             return 1
