@@ -1,17 +1,20 @@
-from dql.vendored.adlfs import AzureBlobFileSystem
+from adlfs import AzureBlobFileSystem
 
-from .fsspec import FSSpecClient
+from .fsspec import DELIMITER, Client
 
 
-class AzureClient(FSSpecClient):
+class AzureClient(Client):
     FS_CLASS = AzureBlobFileSystem
     PREFIX = "az://"
     protocol = "az"
 
-    def _dict_from_info(self, v, parent_id, delimiter, path):
+    def url(self, path: str, expires: int = 3600) -> str:
+        return self.fs.url(path, expires=expires)
+
+    def _dict_from_info(self, v, parent_id, parent, partial_id):
         version_id = v.get("version_id")
         checksum = v.get("content_settings", {}).get("content_md5", "")
-        name = v.get("name", "").split(delimiter)[-1]
+        name = v.get("name", "").split(DELIMITER)[-1]
         if version_id:
             version_suffix = f"?versionid={version_id}"
             if name.endswith(version_suffix):
@@ -19,7 +22,7 @@ class AzureClient(FSSpecClient):
         return {
             "dir_id": None,
             "parent_id": parent_id,
-            "path": path,
+            "parent": parent,
             "name": name,
             "checksum": checksum or "",
             "etag": v.get("etag", ""),
@@ -30,4 +33,5 @@ class AzureClient(FSSpecClient):
             "owner_name": "",
             "owner_id": "",
             "anno": None,
+            "partial_id": partial_id,
         }
