@@ -4,18 +4,22 @@ from datetime import datetime
 from dateutil.parser import isoparse
 from gcsfs import GCSFileSystem
 
-from .fsspec import FSSpecClient
+from .fsspec import DELIMITER, Client
 
 # Patch gcsfs for consistency with s3fs
 # pylint:disable-next=protected-access
 GCSFileSystem.set_session = GCSFileSystem._set_session
 
 
-class GCSClient(FSSpecClient):
+class GCSClient(Client):
     FS_CLASS = GCSFileSystem
     PREFIX = "gcs://"
     protocol = "gcs"
 
+    def url(self, path: str, expires: int = 3600) -> str:
+        # TODO https://github.com/iterative/dql/issues/361
+        raise NotImplementedError("Signed urls are not implemented for gcs")
+
     @staticmethod
     def parse_timestamp(timestamp: str) -> datetime:
         """
@@ -27,8 +31,8 @@ class GCSClient(FSSpecClient):
         assert dt.tzinfo is not None
         return dt
 
-    def _dict_from_info(self, v, parent_id, delimiter, path):
-        name = v.get("name", "").split(delimiter)[-1]
+    def _dict_from_info(self, v, parent_id, parent, partial_id):
+        name = v.get("name", "").split(DELIMITER)[-1]
         if "generation" in v:
             gen = f"#{v['generation']}"
             if name.endswith(gen):
@@ -36,7 +40,7 @@ class GCSClient(FSSpecClient):
         return {
             "is_dir": False,
             "parent_id": parent_id,
-            "path": path,
+            "parent": parent,
             "name": name,
             # 'expires': expires,
             "checksum": b64decode(v.get("md5Hash", "")).hex(),
@@ -49,4 +53,5 @@ class GCSClient(FSSpecClient):
             "owner_name": "",
             "owner_id": "",
             "anno": None,
+            "partial_id": partial_id,
         }
