@@ -1,36 +1,33 @@
-import hashlib
 import os
 import posixpath
-import sys
 from datetime import datetime, timezone
-from functools import partial
 from pathlib import Path
 from typing import Tuple
 
 from fsspec.implementations.local import LocalFileSystem
 
 from dql.data_storage import AbstractDataStorage
+from dql.error import StorageNotFoundError
 from dql.nodes_fetcher import NodesFetcher
 
-from .fsspec import FSSpecClient
-
-if sys.version_info < (3, 9):
-    md5 = hashlib.md5
-else:
-    md5 = partial(hashlib.md5, usedforsecurity=False)
+from .checksum import file_digest
+from .fsspec import Client
 
 
-class FileClient(FSSpecClient):
+class FileClient(Client):
     FS_CLASS = LocalFileSystem
     PREFIX = "file://"
     protocol = "file"
 
     def __init__(
-        self, name: str, fs: LocalFileSystem, use_symlinks: bool = False
+        self, name: str, fs: LocalFileSystem, cache, use_symlinks: bool = False
     ) -> None:
-        super().__init__(name, fs)
+        super().__init__(name, fs, cache)
         self.use_symlinks = use_symlinks
 
+    def url(self, path: str, expires: int = 3600) -> str:
+        raise TypeError("Signed urls are not implemented for local file system")
+
     @classmethod
     def ls_buckets(cls, **kwargs):
         return []
@@ -38,16 +35,26 @@ class FileClient(FSSpecClient):
     @classmethod
     def split_url(cls, url: str, data_storage) -> Tuple[str, str]:
         # pylint:disable=protected-access
-        if data_storage.get_storage(url):
+        def _storage_exists(uri: str) -> bool:
+            try:
+                data_storage.get_storage(uri)
+            except StorageNotFoundError:
+                return False
+            return True
+
+        # lowercasing scheme just in case it's uppercase
+        scheme, rest = url.split(":", 1)
+        url = f"{scheme.lower()}:{rest}"
+        if _storage_exists(url):
             return LocalFileSystem._strip_protocol(url), ""
         for pos in range(len(url) - 1, len(cls.PREFIX), -1):
-            if url[pos] == "/" and data_storage.get_storage(url[:pos]):
+            if url[pos] == "/" and _storage_exists(url[:pos]):
                 return LocalFileSystem._strip_protocol(url[:pos]), url[pos + 1 :]
         raise RuntimeError(f"Invalid file path '{url}'")
 
     @classmethod
-    def from_url(cls, url: str, data_storage, kwargs):
-        result = cls(url, cls.create_fs(**kwargs))
+    def from_url(cls, url: str, data_storage, cache, kwargs):
+        result = cls(url, cls.create_fs(**kwargs), cache)
         storage = data_storage.get_storage(result.uri)
         result.use_symlinks = storage.symlinks
         return result
@@ -68,12 +75,12 @@ class FileClient(FSSpecClient):
             full_path += "/"
         return full_path
 
-    def _dict_from_info(self, v, parent_id, delimiter, path):
-        name = posixpath.basename(path)
+    def _dict_from_info(self, v, parent_id, parent, partial_id):
+        name = posixpath.basename(v["name"])
         return {
             "dir_id": None,
             "parent_id": parent_id,
-            "path": self.rel_path(v["name"]),
+            "parent": parent,
             "name": name,
             "checksum": "",
             "etag": "",
@@ -84,13 +91,13 @@ class FileClient(FSSpecClient):
             "owner_name": "",
             "owner_id": "",
             "anno": None,
+            "partial_id": partial_id,
         }
 
     def fetch_nodes(
         self,
         file_path,
         nodes,
-        cache,
         data_storage: AbstractDataStorage,
         total_size=None,
         cls=NodesFetcher,
@@ -103,7 +110,6 @@ class FileClient(FSSpecClient):
             return super().fetch_nodes(
                 file_path,
                 nodes,
-                cache,
                 data_storage,
                 total_size,
                 cls,
@@ -113,28 +119,19 @@ class FileClient(FSSpecClient):
 
     def _fetch_symlinks(self, nodes, data_storage):
         updated_nodes = []
-        BUFSIZE = 2**18
-        buf = bytearray(BUFSIZE)  # Reusable buffer to reduce allocations.
-        view = memoryview(buf)
         for node in nodes:
             if not node.is_downloadable:
                 continue
-            src_path = f"{self.name}/{node.path_str}"
-            digestobj = md5()
+            src_path = f"{self.name}/{node.path}"
             with open(src_path, "rb") as fileobj:
-                # From 3.11's hashlib.filedigest()
-                while True:
-                    size = fileobj.readinto(buf)
-                    if size == 0:
-                        break  # EOF
-                    digestobj.update(view[:size])
-            md5_hash = digestobj.hexdigest()
+                md5_hash = file_digest(fileobj)
             data_storage.update_checksum(node, md5_hash)
-            updated_nodes.append(node._replace(checksum=md5_hash))
+            node.checksum = md5_hash
+            updated_nodes.append(node)
         return updated_nodes
 
-    def do_instantiate_node(self, node, cache, dst):
+    def do_instantiate_node(self, node, dst):
         if self.use_symlinks:
-            os.symlink(Path(self.name, node.path_str), dst)
+            os.symlink(Path(self.name, node.path), dst)
         else:
-            super().do_instantiate_node(node, cache, dst)
+            super().do_instantiate_node(node, dst)
