@@ -17,9 +17,10 @@ class NodesFetcher(NodesThreadPool):
         updated_nodes = []
         for d in done:
             lst = d.result()
-            for node, md5 in lst:
-                self.data_storage.update_checksum(node, md5)
-                updated_nodes.append(node._replace(checksum=md5))
+            for node, checksum in lst:
+                self.data_storage.update_checksum(node, checksum)
+                node.checksum = checksum
+                updated_nodes.append(node)
         return updated_nodes
 
     def do_task(self, chunk):
@@ -29,12 +30,11 @@ class NodesFetcher(NodesThreadPool):
                 self.increase_counter(node.size)
                 continue
 
-            pair = self.fetch(self.client.name, node.path_str, node)
+            pair = self.fetch(node.path, node)
             res.append(pair)
         return res
 
-    def fetch(self, bucket, path, node):
-        from dvc_data.hashfile.build import _upload_file
+    def fetch(self, path, node):
         from dvc_objects.fs.callbacks import Callback
 
         class _CB(Callback):
@@ -43,26 +43,7 @@ class NodesFetcher(NodesThreadPool):
             ):
                 self.increase_counter(inc)
 
-        _, obj = _upload_file(
-            f"{bucket}/{path}",
-            self.client.fs,
-            self.cache,
-            self.cache,
-            callback=_CB(),
+        hash_value = self.client.download(
+            path, node.vtype, node.location, callback=_CB()
         )
-
-        return node, obj.hash_info.value
-
-
-class JsonNodesFetcher(NodesFetcher):
-    def fetch(self, bucket, path, node):
-        node, md5 = super().fetch(bucket, path, node)
-
-        md5_fname = self.cache.oid_to_path(md5)
-        with open(md5_fname, encoding="utf-8") as fd:
-            content = fd.read()
-            content_escape_quotes = content.replace("'", "''")
-
-        self.data_storage.update_annotation(node, content_escape_quotes)
-
-        return node, md5
+        return node, hash_value
