@@ -75,8 +75,7 @@ class NodesThreadPool(ABC):
                 total=total_size,
             )
 
-            chunk = next(chunk_gen, None)
-            while chunk:
+            for chunk in chunk_gen:
                 while len(tasks) >= self._max_threads:
                     done, _ = concurrent.futures.wait(
                         tasks, timeout=1, return_when="FIRST_COMPLETED"
@@ -87,11 +86,8 @@ class NodesThreadPool(ABC):
                     self.update_progress_bar(progress_bar)
 
                 tasks.add(thPool.submit(self.do_task, chunk))
-
                 self.update_progress_bar(progress_bar)
 
-                chunk = next(chunk_gen, None)
-
             while tasks:
                 done, _ = concurrent.futures.wait(
                     tasks, timeout=1, return_when="FIRST_COMPLETED"
