@@ -1,8 +1,19 @@
+import importlib.util
 import os
 import os.path as osp
 import stat
+import sys
 from datetime import datetime
-from typing import TYPE_CHECKING, Iterable, Mapping, Optional, Type, TypeVar, Union
+from typing import (
+    TYPE_CHECKING,
+    Dict,
+    Iterable,
+    Mapping,
+    Optional,
+    Type,
+    TypeVar,
+    Union,
+)
 
 from dateutil import tz
 from dateutil.parser import isoparse
@@ -12,6 +23,7 @@ if TYPE_CHECKING:
     from tomlkit import TOMLDocument
 
 GLOB_CHARS = ["?", "*", "[", "]"]
+NUL = b"\0"
 
 T = TypeVar("T", bound="DQLDir")
 
@@ -22,6 +34,7 @@ class DQLDir:
     TMP = "tmp"
     DB = "db"
     ENV_VAR = "DQL_DIR"
+    ENV_VAR_DQL_ROOT = "DQL_ROOT_DIR"
 
     def __init__(
         self,
@@ -47,7 +60,12 @@ class DQLDir:
 
     @classmethod
     def default_root(cls) -> str:
-        return osp.join(os.getcwd(), cls.DEFAULT)
+        try:
+            root_dir = os.environ[cls.ENV_VAR_DQL_ROOT]
+        except KeyError:
+            root_dir = os.getcwd()
+
+        return osp.join(root_dir, cls.DEFAULT)
 
     @classmethod
     def find(cls: Type[T], create: bool = True) -> T:
@@ -209,3 +227,45 @@ def remove_readonly(func, path, _):
     "Clear the readonly bit and reattempt the removal"
     os.chmod(path, stat.S_IWRITE)
     func(path)
+
+
+def sql_escape_like(search: str, escape: str = "\\") -> str:
+    return (
+        search.replace(escape, escape * 2)
+        .replace("%", f"{escape}%")
+        .replace("_", f"{escape}_")
+    )
+
+
+def get_envs_by_prefix(prefix: str) -> Dict[str, str]:
+    """
+    Function that searches env variables by some name prefix and returns
+    the ones found, but with prefix being excluded from it's names
+    """
+    variables: Dict[str, str] = {}
+    for env_name, env_value in os.environ.items():
+        if env_name.startswith(prefix):
+            variables[env_name[len(prefix) :]] = env_value
+
+    return variables
+
+
+def import_object(object_spec):
+    filename, identifier = object_spec.rsplit(":", 1)
+    filename = filename.strip()
+    identifier = identifier.strip()
+
+    if not identifier.isidentifier() or not filename.endswith(".py"):
+        raise ValueError(f"Invalid object spec: {object_spec}")
+
+    modname = os.path.abspath(filename)
+    if modname in sys.modules:
+        module = sys.modules[modname]
+    else:
+        # Use importlib to find and load the module from the given filename
+        spec = importlib.util.spec_from_file_location(modname, filename)
+        module = importlib.util.module_from_spec(spec)
+        sys.modules[modname] = module
+        spec.loader.exec_module(module)
+
+    return getattr(module, identifier)
