@@ -1,4 +1,5 @@
 import os
+import uuid
 from pathlib import PosixPath
 from typing import Dict, Generator, Type
 
@@ -8,10 +9,13 @@ from pytest import MonkeyPatch, TempPathFactory
 from upath.implementations.cloud import CloudPath
 
 from dql.catalog import Catalog
+from dql.catalog.loader import get_db_adapter
 from dql.data_storage import SQLiteDataStorage
 from dql.utils import DQLDir
 
-# pylint: disable=redefined-outer-name
+from .utils import DEFAULT_TREE, instantiate_tree
+
+# pylint: disable=redefined-outer-name,unused-argument
 
 DEFAULT_DQL_BIN = "dql"
 DEFAULT_DQL_GIT_REPO = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
@@ -53,18 +57,24 @@ def clean_environment(
 
 @pytest.fixture
 def data_storage():
-    _data_storage = SQLiteDataStorage(db_file=":memory:")
-    yield _data_storage
+    if os.environ.get("DQL_DB_ADAPTER"):
+        _data_storage = get_db_adapter()
+        yield _data_storage
 
-    # Wipe the DB after the test
-    # Using new connection to check that the DB isn't locked
-    conn = _data_storage.clone().db
-    tables = conn.execute(
-        "SELECT name FROM sqlite_master WHERE type='table'"
-    ).fetchall()
-    for (table,) in tables:
-        conn.execute(f"DROP TABLE '{table}'")
-    conn.commit()
+        _data_storage.cleanup_for_tests()
+    else:
+        _data_storage = SQLiteDataStorage(db_file=":memory:")
+        yield _data_storage
+
+        # Wipe the DB after the test
+        # Using new connection to check that the DB isn't locked
+        conn = _data_storage.clone().db
+        tables = conn.execute(
+            "SELECT name FROM sqlite_master WHERE type='table'"
+        ).fetchall()
+        for (table,) in tables:
+            conn.execute(f"DROP TABLE '{table}'")
+        conn.commit()
 
 
 @pytest.fixture
@@ -124,6 +134,11 @@ def pytest_configure(config):
     config.dql_config.dql_git_repo = config.getoption("--dql-git-repo")
 
 
+@pytest.fixture(scope="session", params=[DEFAULT_TREE])
+def tree(request):
+    return request.param
+
+
 @attrs.define
 class CloudServer:
     kind: str
@@ -131,7 +146,7 @@ class CloudServer:
     client_config: Dict[str, str]
 
 
-def make_cloud_server(src_path, cloud_type):
+def make_cloud_server(src_path, cloud_type, tree):
     fs = src_path.fs
     if cloud_type == "s3":
         endpoint_url = fs.client_kwargs["endpoint_url"]
@@ -145,17 +160,8 @@ def make_cloud_server(src_path, cloud_type):
         client_config = {}
     else:
         raise ValueError(f"invalid cloud_type: {cloud_type}")
-    (src_path / "description").write_text("Cats and Dogs")
-    (src_path / "cats").mkdir()
-    (src_path / "cats" / "cat1").write_text("meow")
-    (src_path / "cats" / "cat2").write_text("mrow")
-    (src_path / "dogs").mkdir()
-    (src_path / "dogs" / "dog1").write_text("woof")
-    (src_path / "dogs" / "dog2").write_text("arf")
-    (src_path / "dogs" / "dog3").write_text("bark")
-    (src_path / "dogs" / "others").mkdir()
-    (src_path / "dogs" / "others" / "dog4").write_text("ruff")
 
+    instantiate_tree(src_path, tree)
     return CloudServer(kind=cloud_type, src=src_path, client_config=client_config)
 
 
@@ -191,7 +197,7 @@ def version_aware(request):
 
 
 @pytest.fixture(scope="session")
-def cloud_server(request, tmp_upath_factory, cloud_type, version_aware):
+def cloud_server(request, tmp_upath_factory, cloud_type, version_aware, tree):
     if cloud_type == "azure" and version_aware:
         if conn_str := request.config.getoption("--azure-connection-string"):
             src_path = tmp_upath_factory.azure(conn_str)
@@ -204,24 +210,56 @@ def cloud_server(request, tmp_upath_factory, cloud_type, version_aware):
             src_path = tmp_upath_factory.mktemp("local")
     else:
         src_path = tmp_upath_factory.mktemp(cloud_type, version_aware=version_aware)
-    return make_cloud_server(src_path, cloud_type)
+    return make_cloud_server(src_path, cloud_type, tree)
 
 
 @pytest.fixture
 def cloud_test_catalog(
     cloud_server,
     tmp_path,
-    data_storage,  # pylint: disable=redefined-outer-name # noqa: E501
+    data_storage,
 ):
     cache_dir = tmp_path / ".dql" / "cache"
     cache_dir.mkdir(parents=True)
     tmpfile_dir = tmp_path / ".dql" / "tmp"
     tmpfile_dir.mkdir()
 
-    catalog = Catalog(data_storage, cache_dir=str(cache_dir), tmp_dir=str(tmpfile_dir))
+    catalog = Catalog(
+        data_storage,
+        cache_dir=str(cache_dir),
+        tmp_dir=str(tmpfile_dir),
+        client_config=cloud_server.client_config,
+    )
     result = CloudTestCatalog(
         server=cloud_server, working_dir=tmp_path, catalog=catalog
     )
     if cloud_server.kind == "file":
         catalog.add_storage(f"{result.src_uri}")
     return result
+
+
+@pytest.fixture
+def listed_bucket(cloud_test_catalog):
+    list(
+        cloud_test_catalog.catalog.ls(
+            [cloud_test_catalog.src_uri],
+            fields=["name"],
+            client_config=cloud_test_catalog.client_config,
+        )
+    )
+
+
+@pytest.fixture
+def dogs_shadow_dataset(listed_bucket, cloud_test_catalog):
+    shadow_dataset_name = uuid.uuid4().hex
+    src_uri = cloud_test_catalog.src_uri
+    catalog = cloud_test_catalog.catalog
+
+    catalog.create_shadow_dataset(
+        shadow_dataset_name,
+        [f"{src_uri}/dogs/*"],
+        client_config=cloud_test_catalog.client_config,
+        recursive=True,
+    )
+
+    return catalog.data_storage.get_dataset(shadow_dataset_name)
