@@ -22,24 +22,14 @@ def nodes():
 
 
 def make_size_node(node_id, dir_type, parent_id, name, size):
-    return Node._make(
-        (
-            node_id,
-            dir_type,
-            parent_id,
-            name,
-            "",
-            "",
-            "",
-            True,
-            None,
-            size,
-            "",
-            "",
-            name,
-            None,
-            True,
-        )
+    return Node(
+        node_id,
+        "",  # vtype
+        dir_type,
+        parent_id,
+        "",  # parent
+        name,
+        size=size,
     )
 
 
