@@ -2,11 +2,11 @@ from sqlalchemy import JSON, Column, Float
 from sqlalchemy.dialects.sqlite import dialect as sqlite_dialect
 from sqlalchemy.schema import CreateTable
 
-from dql.dataset import dataset_table
+from dql.data_storage.schema import DatasetRow
 
 
 def test_dataset_table_compilation():
-    table = dataset_table(
+    table = DatasetRow.new_table(
         "ds-1",
         custom_columns=[
             Column("score", Float, nullable=False),
@@ -14,12 +14,15 @@ def test_dataset_table_compilation():
         ],
     )
     result = CreateTable(table, if_not_exists=True).compile(dialect=sqlite_dialect())
+
     assert result.string == (
         "\n"
         'CREATE TABLE IF NOT EXISTS "ds-1" (\n'
         "\tid INTEGER NOT NULL, \n"
+        "\tvtype TEXT NOT NULL, \n"
         "\tdir_type INTEGER, \n"
         "\tparent_id INTEGER, \n"
+        "\tparent TEXT, \n"
         "\tname TEXT NOT NULL, \n"
         "\tchecksum TEXT, \n"
         "\tetag TEXT, \n"
@@ -29,13 +32,13 @@ def test_dataset_table_compilation():
         "\tsize BIGINT NOT NULL, \n"
         "\towner_name TEXT, \n"
         "\towner_id TEXT, \n"
-        "\tpath_str TEXT, \n"
         "\tanno JSON, \n"
+        "\trandom BIGINT NOT NULL, \n"
+        "\tlocation JSON, \n"
         "\tsource TEXT NOT NULL, \n"
         "\tscore FLOAT NOT NULL, \n"
         "\tmeta_info JSON, \n"
-        "\tPRIMARY KEY (id), \n"
-        "\tUNIQUE (source, path_str)\n"
+        "\tPRIMARY KEY (id)\n"
         ")\n"
         "\n"
     )
