@@ -4,10 +4,13 @@ from datetime import datetime
 import pytest
 
 from dql.catalog import Catalog
+from dql.catalog.catalog import DataSource
 from dql.data_storage import SQLiteDataStorage
 from dql.listing import Listing
 from dql.node import DirType
 
+from ..utils import skip_if_not_sqlite
+
 # pylint: disable=redefined-outer-name
 
 TS = datetime(2022, 8, 1)
@@ -27,11 +30,11 @@ async def _tree_to_db(lst: Listing, tree: dict, dir_id: int = 1, path=""):
     for k, v in tree.items():
         if k:
             dir_path = posixpath.join(path, k)
-            new_dir_id = await lst.insert_dir(dir_id, k, TS, posixpath.join(path, k))
+            new_dir_id = await lst.insert_dir(dir_id, k, TS, path, 0)
             await _tree_to_db(lst, v, new_dir_id, dir_path)
         else:
             for fname in v:
-                await lst.insert_file(dir_id, fname, TS, posixpath.join(path, fname))
+                await lst.insert_file(dir_id, fname, TS, path, 0)
 
 
 @pytest.fixture
@@ -44,7 +47,8 @@ def data_storage():
 @pytest.fixture
 async def listing(data_storage):
     bkt = "s3://whatever"
-    lst, _ = Catalog(data_storage).enlist_source(bkt, 1234, to_fetch=False)
+    lst, _ = Catalog(data_storage).enlist_source(bkt, 1234, skip_indexing=True)
+    lst.data_storage.init_db()
     await _tree_to_db(lst, TREE, await lst.insert_root())
     return lst
 
@@ -75,7 +79,7 @@ async def test_resolve_not_existing_path(listing):
 
 async def test_resolve_root(listing):
     node = listing.resolve_path("")
-    assert node.dir_type == DirType.ROOT
+    assert node.dir_type == DirType.DIR
     assert node.is_dir
     assert node.name == ""
 
@@ -102,7 +106,7 @@ async def test_file_ends_with_slash(listing):
 def _match_filenames(nodes, expected_names):
     assert len(nodes) == len(expected_names)
     names = (node.name for node in nodes)
-    assert set(names).intersection(set(expected_names))
+    assert set(names) == set(expected_names)
 
 
 async def test_basic_expansion(listing):
@@ -116,6 +120,7 @@ async def test_subname_expansion(listing):
 
 
 async def test_multilevel_expansion(listing):
+    skip_if_not_sqlite()
     nodes = listing.expand_path("dir[1,2]/d*")
     _match_filenames(nodes, ["dataset.csv", "diagram.png", "d2"])
 
@@ -123,28 +128,28 @@ async def test_multilevel_expansion(listing):
 async def test_expand_root(listing):
     nodes = listing.expand_path("")
     assert len(nodes) == 1
-    assert nodes[0].dir_type == DirType.ROOT
+    assert nodes[0].dir_type == DirType.DIR
     assert nodes[0].is_dir
 
 
 async def test_list_dir(listing):
     dir1 = listing.resolve_path("dir1/")
-    nodes = list(listing.ls_path(dir1))
-    _match_filenames(nodes, ["dir2", "dataset.csv"])
+    names = listing.ls_path(dir1, ["name"])
+    assert {n[0] for n in names} == {"d2", "dataset.csv"}
 
 
 async def test_list_file(listing):
     file = listing.resolve_path("dir1/dataset.csv")
-    nodes = list(listing.ls_path(file))
-    _match_filenames(nodes, ["dataset.csv"])
-    assert nodes[0].id == file.id
-    assert nodes[0].dir_type == DirType.FILE
-    assert not nodes[0].is_dir
-    assert nodes[0].name == file.name
+    src = DataSource(listing, file)
+    results = list(src.ls(["id", "name", "dir_type"]))
+    assert {r[1] for r in results} == {"dataset.csv"}
+    assert results[0][0] == file.id
+    assert results[0][1] == file.name
+    assert results[0][2] == DirType.FILE
 
 
 async def test_subtree(listing):
     dir1 = listing.resolve_path("dir1/")
-    nodes = listing.data_storage.walk_subtree(dir1)
-    subtree_files = ["dir2", "dataset.csv", "file1.csv", "file2.csv"]
-    _match_filenames(list(nodes), subtree_files)
+    nodes = listing.data_storage.get_subtree_files(dir1)
+    subtree_files = ["dataset.csv", "file1.csv", "file2.csv"]
+    _match_filenames([nwp.n for nwp in nodes], subtree_files)
