@@ -1,8 +1,10 @@
 from datetime import datetime, timedelta, timezone
 
 import pytest
+from sqlalchemy import func
 
 from dql import utils
+from dql.error import StorageNotFoundError
 from dql.storage import STALE_HOURS_LIMIT, Status, Storage
 
 TS = datetime(2022, 8, 1)
@@ -30,23 +32,31 @@ def test_expiration_time():
 
 
 def test_adding_storage(data_storage):
-    # These tests re-use the same database for better performance
-    data_storage.db.execute("DELETE FROM buckets")
-    data_storage.db.commit()
     uri = "s3://whatever"
-    cached_storage = data_storage.get_storage(uri)
-    assert cached_storage is None
+    with pytest.raises(StorageNotFoundError):
+        data_storage.get_storage(uri)
 
     storage, _, _, _ = data_storage.register_storage_for_indexing(uri)
-    cnt = data_storage.db.execute("SELECT COUNT() FROM buckets").fetchall()
-    assert cnt[0][0] == 1
+    cnt = next(data_storage.execute(data_storage.storages_select(func.count())), (0,))
+    assert cnt[0] == 1
 
-    query = "SELECT * FROM buckets WHERE uri = ?"
-    bkt = data_storage.db.execute(query, [uri]).fetchall()
+    bkt = list(
+        data_storage.execute(
+            data_storage.storages_select().where(data_storage.storages.c.uri == uri)
+        )
+    )
     assert len(bkt) == 1
 
-    s = Storage(*bkt[0])
-    assert s == storage
+    assert storage.id == bkt[0][0]
+    assert storage.uri == bkt[0][1]
+    assert storage.timestamp == bkt[0][2]
+    assert storage.expires == bkt[0][3]
+    assert storage.started_inserting_at == bkt[0][4]
+    assert storage.last_inserted_at == bkt[0][5]
+    assert storage.status == bkt[0][6]
+    assert storage.symlinks == bkt[0][7]
+    assert storage.error_message == ""
+    assert storage.error_stack == ""
 
 
 def test_storage_status(data_storage):
@@ -92,7 +102,7 @@ def test_storage_status(data_storage):
 def test_max_ttl(ttl):
     uri = "s3://whatever"
     expires = Storage.get_expiration_time(TS, ttl)
-    storage = Storage(uri, TS, expires)
+    storage = Storage(1, uri, TS, expires)
     assert storage.timestamp == TS
     assert storage.expires == datetime.max
     assert storage.timestamp_str  # no error
@@ -102,7 +112,7 @@ def test_max_ttl(ttl):
 
 def test_storage_without_dates():
     uri = "s3://whatever"
-    storage = Storage(uri, None, None)
+    storage = Storage(1, uri, None, None)
     assert storage.timestamp is None
     assert storage.expires is None
     assert storage.timestamp_str is None  # no error
@@ -138,7 +148,8 @@ def test_stale_storage(data_storage):
         "last_inserted_at": datetime.now(timezone.utc)
         - timedelta(hours=STALE_HOURS_LIMIT + 1)
     }
-    data_storage.storages.where(uri=uri_stale).update(**updates)
+    s = data_storage.storages
+    data_storage.execute(s.update().where(s.c.uri == uri_stale).values(**updates))
 
     data_storage.find_stale_storages()
 
@@ -147,3 +158,25 @@ def test_stale_storage(data_storage):
 
     not_stale_storage = data_storage.get_storage(uri_not_stale)
     assert not_stale_storage.status == Status.PENDING
+
+
+def test_failed_storage(data_storage):
+    uri = "s3://bucket"
+    error_message = "Internal error on indexing"
+    error_stack = "error"
+    data_storage.create_storage_if_not_registered(uri)
+
+    data_storage.mark_storage_pending(data_storage.get_storage(uri))
+    data_storage.mark_storage_indexed(
+        uri,
+        Status.FAILED,
+        1000,
+        datetime.now(),
+        error_message=error_message,
+        error_stack=error_stack,
+    )
+
+    storage = data_storage.get_storage(uri)
+    assert storage.status == Status.FAILED
+    assert storage.error_message == error_message
+    assert storage.error_stack == error_stack
